#!/usr/bin/python
# -*- coding: utf-8 -*-

# Copyright: Ansible Project
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import absolute_import, division, print_function
__metaclass__ = type


DOCUMENTATION = '''
---
module: ec2_eip
version_added: 1.0.0
short_description: manages EC2 elastic IP (EIP) addresses.
description:
    - This module can allocate or release an EIP.
    - This module can associate/disassociate an EIP with instances or network interfaces.
options:
  device_id:
    description:
      - The id of the device for the EIP. Can be an EC2 Instance id or Elastic Network Interface (ENI) id.
    required: false
    aliases: [ instance_id ]
    type: str
  public_ip:
    description:
      - The IP address of a previously allocated EIP.
      - When I(public_ip=present) and device is specified, the EIP is associated with the device.
      - When I(public_ip=absent) and device is specified, the EIP is disassociated from the device.
    aliases: [ ip ]
    type: str
  state:
    description:
      - When C(state=present), allocate an EIP or associate an existing EIP with a device.
      - When C(state=absent), disassociate the EIP from the device and optionally release it.
    choices: ['present', 'absent']
    default: present
    type: str
  in_vpc:
    description:
      - Allocate an EIP inside a VPC or not.
      - Required if specifying an ENI with I(device_id).
    default: false
    type: bool
  reuse_existing_ip_allowed:
    description:
      - Reuse an EIP that is not associated to a device (when available), instead of allocating a new one.
    default: false
    type: bool
  release_on_disassociation:
    description:
      - Whether or not to automatically release the EIP when it is disassociated.
    default: false
    type: bool
  private_ip_address:
    description:
      - The primary or secondary private IP address to associate with the Elastic IP address.
    type: str
  allow_reassociation:
    description:
      -  Specify this option to allow an Elastic IP address that is already associated with another
         network interface or instance to be re-associated with the specified instance or interface.
    default: false
    type: bool
  tag_name:
    description:
      - When I(reuse_existing_ip_allowed=true), supplement with this option to only reuse
        an Elastic IP if it is tagged with I(tag_name).
    type: str
  tag_value:
    description:
      - Supplements I(tag_name) but also checks that the value of the tag provided in I(tag_name) matches I(tag_value).
    type: str
  public_ipv4_pool:
    description:
      - Allocates the new Elastic IP from the provided public IPv4 pool (BYOIP)
        only applies to newly allocated Elastic IPs, isn't validated when I(reuse_existing_ip_allowed=true).
    type: str
  wait_timeout:
    description:
      - The I(wait_timeout) option does nothing and will be removed after 2022-06-01
    type: int
extends_documentation_fragment:
- amazon.aws.aws
- amazon.aws.ec2

author: "Rick Mendes (@rickmendes) <rmendes@illumina.com>"
notes:
   - There may be a delay between the time the EIP is assigned and when
     the cloud instance is reachable via the new address. Use wait_for and
     pause to delay further playbook execution until the instance is reachable,
     if necessary.
   - This module returns multiple changed statuses on disassociation or release.
     It returns an overall status based on any changes occurring. It also returns
     individual changed statuses for disassociation and release.
'''

EXAMPLES = '''
# Note: These examples do not set authentication details, see the AWS Guide for details.

- name: associate an elastic IP with an instance
  community.aws.ec2_eip:
    device_id: i-1212f003
    ip: 93.184.216.119

- name: associate an elastic IP with a device
  community.aws.ec2_eip:
    device_id: eni-c8ad70f3
    ip: 93.184.216.119

- name: associate an elastic IP with a device and allow reassociation
  community.aws.ec2_eip:
    device_id: eni-c8ad70f3
    public_ip: 93.184.216.119
    allow_reassociation: true

- name: disassociate an elastic IP from an instance
  community.aws.ec2_eip:
    device_id: i-1212f003
    ip: 93.184.216.119
    state: absent

- name: disassociate an elastic IP with a device
  community.aws.ec2_eip:
    device_id: eni-c8ad70f3
    ip: 93.184.216.119
    state: absent

- name: allocate a new elastic IP and associate it with an instance
  community.aws.ec2_eip:
    device_id: i-1212f003

- name: allocate a new elastic IP without associating it to anything
  community.aws.ec2_eip:
    state: present
  register: eip

- name: output the IP
  ansible.builtin.debug:
    msg: "Allocated IP is {{ eip.public_ip }}"

- name: provision new instances with ec2
  amazon.aws.ec2:
    keypair: mykey
    instance_type: c1.medium
    image: ami-40603AD1
    wait: true
    group: webserver
    count: 3
  register: ec2

- name: associate new elastic IPs with each of the instances
  community.aws.ec2_eip:
    device_id: "{{ item }}"
  loop: "{{ ec2.instance_ids }}"

- name: allocate a new elastic IP inside a VPC in us-west-2
  community.aws.ec2_eip:
    region: us-west-2
    in_vpc: true
  register: eip

- name: output the IP
  ansible.builtin.debug:
    msg: "Allocated IP inside a VPC is {{ eip.public_ip }}"

- name: allocate eip - reuse unallocated ips (if found) with FREE tag
  community.aws.ec2_eip:
    region: us-east-1
    in_vpc: true
    reuse_existing_ip_allowed: true
    tag_name: FREE

- name: allocate eip - reuse unallocated ips if tag reserved is nope
  community.aws.ec2_eip:
    region: us-east-1
    in_vpc: true
    reuse_existing_ip_allowed: true
    tag_name: reserved
    tag_value: nope

- name: allocate new eip - from servers given ipv4 pool
  community.aws.ec2_eip:
    region: us-east-1
    in_vpc: true
    public_ipv4_pool: ipv4pool-ec2-0588c9b75a25d1a02

- name: allocate eip - from a given pool (if no free addresses where dev-servers tag is dynamic)
  community.aws.ec2_eip:
    region: us-east-1
    in_vpc: true
    reuse_existing_ip_allowed: true
    tag_name: dev-servers
    public_ipv4_pool: ipv4pool-ec2-0588c9b75a25d1a02

- name: allocate eip from pool - check if tag reserved_for exists and value is our hostname
  community.aws.ec2_eip:
    region: us-east-1
    in_vpc: true
    reuse_existing_ip_allowed: true
    tag_name: reserved_for
    tag_value: "{{ inventory_hostname }}"
    public_ipv4_pool: ipv4pool-ec2-0588c9b75a25d1a02
'''

RETURN = '''
allocation_id:
  description: allocation_id of the elastic ip
  returned: on success
  type: str
  sample: eipalloc-51aa3a6c
public_ip:
  description: an elastic ip address
  returned: on success
  type: str
  sample: 52.88.159.209
'''

try:
    import botocore.exceptions
except ImportError:
    pass  # caught by AnsibleAWSModule

from ansible_collections.amazon.aws.plugins.module_utils.core import AnsibleAWSModule
from ansible_collections.amazon.aws.plugins.module_utils.core import is_boto3_error_code
from ansible_collections.amazon.aws.plugins.module_utils.ec2 import AWSRetry
from ansible_collections.amazon.aws.plugins.module_utils.ec2 import ansible_dict_to_boto3_filter_list


def associate_ip_and_device(ec2, module, address, private_ip_address, device_id, allow_reassociation, check_mode, is_instance=True):
    if address_is_associated_with_device(ec2, module, address, device_id, is_instance):
        return {'changed': False}

    # If we're in check mode, nothing else to do
    if not check_mode:
        if is_instance:
            try:
                params = dict(
                    InstanceId=device_id,
                    AllowReassociation=allow_reassociation,
                )
                if private_ip_address:
                    params['PrivateIPAddress'] = private_ip_address
                if address['Domain'] == 'vpc':
                    params['AllocationId'] = address['AllocationId']
                else:
                    params['PublicIp'] = address['PublicIp']
                res = ec2.associate_address(**params)
            except (botocore.exceptions.BotoCoreError, botocore.exceptions.ClientError) as e:
                msg = "Couldn't associate Elastic IP address with instance '{0}'".format(device_id)
                module.fail_json_aws(e, msg=msg)
        else:
            params = dict(
                NetworkInterfaceId=device_id,
                AllocationId=address['AllocationId'],
                AllowReassociation=allow_reassociation,
            )

            if private_ip_address:
                params['PrivateIpAddress'] = private_ip_address

            try:
                res = ec2.associate_address(aws_retry=True, **params)
            except (botocore.exceptions.BotoCoreError, botocore.exceptions.ClientError) as e:
                msg = "Couldn't associate Elastic IP address with network interface '{0}'".format(device_id)
                module.fail_json_aws(e, msg=msg)
        if not res:
            module.fail_json_aws(e, msg='Association failed.')

    return {'changed': True}


def disassociate_ip_and_device(ec2, module, address, device_id, check_mode, is_instance=True):
    if not address_is_associated_with_device(ec2, module, address, device_id, is_instance):
        return {'changed': False}

    # If we're in check mode, nothing else to do
    if not check_mode:
        try:
            if address['Domain'] == 'vpc':
                res = ec2.disassociate_address(
                    AssociationId=address['AssociationId'], aws_retry=True
                )
            else:
                res = ec2.disassociate_address(
                    PublicIp=address['PublicIp'], aws_retry=True
                )
        except (botocore.exceptions.BotoCoreError, botocore.exceptions.ClientError) as e:
            module.fail_json_aws(e, msg="Dissassociation of Elastic IP failed")

    return {'changed': True}


@AWSRetry.jittered_backoff()
def find_address(ec2, module, public_ip, device_id, is_instance=True):
    """ Find an existing Elastic IP address """
    filters = []
    kwargs = {}

    if public_ip:
        kwargs["PublicIps"] = [public_ip]
    elif device_id:
        if is_instance:
            filters.append({"Name": 'instance-id', "Values": [device_id]})
        else:
            filters.append({'Name': 'network-interface-id', "Values": [device_id]})

    if len(filters) > 0:
        kwargs["Filters"] = filters
    elif len(filters) == 0 and public_ip is None:
        return None

    try:
        addresses = ec2.describe_addresses(**kwargs)
    except is_boto3_error_code('InvalidAddress.NotFound') as e:
        # If we're releasing and we can't find it, it's already gone...
        if module.params.get('state') == 'absent':
            module.exit_json(changed=False)
        module.fail_json_aws(e, msg="Couldn't obtain list of existing Elastic IP addresses")

    addresses = addresses["Addresses"]
    if len(addresses) == 1:
        return addresses[0]
    elif len(addresses) > 1:
        msg = "Found more than one address using args {0}".format(kwargs)
        msg += "Addresses found: {0}".format(addresses)
        module.fail_json_aws(botocore.exceptions.ClientError, msg=msg)


def address_is_associated_with_device(ec2, module, address, device_id, is_instance=True):
    """ Check if the elastic IP is currently associated with the device """
    address = find_address(ec2, module, address["PublicIp"], device_id, is_instance)
    if address:
        if is_instance:
            if "InstanceId" in address and address["InstanceId"] == device_id:
                return address
        else:
            if "NetworkInterfaceId" in address and address["NetworkInterfaceId"] == device_id:
                return address
    return False


def allocate_address(ec2, module, domain, reuse_existing_ip_allowed, check_mode, tag_dict=None, public_ipv4_pool=None):
    """ Allocate a new elastic IP address (when needed) and return it """
    if reuse_existing_ip_allowed:
        filters = []
        if not domain:
            domain = 'standard'
        filters.append({'Name': 'domain', "Values": [domain]})

        if tag_dict is not None:
            filters += ansible_dict_to_boto3_filter_list(tag_dict)

        try:
            all_addresses = ec2.describe_addresses(Filters=filters, aws_retry=True)
        except (botocore.exceptions.BotoCoreError, botocore.exceptions.ClientError) as e:
            module.fail_json_aws(e, msg="Couldn't obtain list of existing Elastic IP addresses")

        all_addresses = all_addresses["Addresses"]

        if domain == 'vpc':
            unassociated_addresses = [a for a in all_addresses
                                      if not a.get('AssociationId', None)]
        else:
            unassociated_addresses = [a for a in all_addresses
                                      if not a['InstanceId']]
        if unassociated_addresses:
            return unassociated_addresses[0], False

    if public_ipv4_pool:
        return allocate_address_from_pool(ec2, module, domain, check_mode, public_ipv4_pool), True

    try:
        result = ec2.allocate_address(Domain=domain, aws_retry=True), True
    except (botocore.exceptions.BotoCoreError, botocore.exceptions.ClientError) as e:
        module.fail_json_aws(e, msg="Couldn't allocate Elastic IP address")
    return result


def release_address(ec2, module, address, check_mode):
    """ Release a previously allocated elastic IP address """

    # If we're in check mode, nothing else to do
    if not check_mode:
        try:
            result = ec2.release_address(AllocationId=address['AllocationId'], aws_retry=True)
        except (botocore.exceptions.BotoCoreError, botocore.exceptions.ClientError) as e:
            module.fail_json_aws(e, msg="Couldn't release Elastic IP address")

    return {'changed': True}


@AWSRetry.jittered_backoff()
def describe_eni_with_backoff(ec2, module, device_id):
    try:
        return ec2.describe_network_interfaces(NetworkInterfaceIds=[device_id])
    except is_boto3_error_code('InvalidNetworkInterfaceID.NotFound') as e:
        module.fail_json_aws(e, msg="Couldn't get list of network interfaces.")


def find_device(ec2, module, device_id, is_instance=True):
    """ Attempt to find the EC2 instance and return it """

    if is_instance:
        try:
            paginator = ec2.get_paginator('describe_instances')
            reservations = list(paginator.paginate(InstanceIds=[device_id]).search('Reservations[]'))
        except (botocore.exceptions.BotoCoreError, botocore.exceptions.ClientError) as e:
            module.fail_json_aws(e, msg="Couldn't get list of instances")

        if len(reservations) == 1:
            instances = reservations[0]['Instances']
            if len(instances) == 1:
                return instances[0]
    else:
        try:
            interfaces = describe_eni_with_backoff(ec2, module, device_id)
        except (botocore.exceptions.BotoCoreError, botocore.exceptions.ClientError) as e:
            module.fail_json_aws(e, msg="Couldn't get list of network interfaces.")
        if len(interfaces) == 1:
            return interfaces[0]


def ensure_present(ec2, module, domain, address, private_ip_address, device_id,
                   reuse_existing_ip_allowed, allow_reassociation, check_mode, is_instance=True):
    changed = False

    # Return the EIP object since we've been given a public IP
    if not address:
        if check_mode:
            return {'changed': True}

        address, changed = allocate_address(ec2, module, domain, reuse_existing_ip_allowed, check_mode)

    if device_id:
        # Allocate an IP for instance since no public_ip was provided
        if is_instance:
            instance = find_device(ec2, module, device_id)
            if reuse_existing_ip_allowed:
                if instance['VpcId'] and len(instance['VpcId']) > 0 and domain is None:
                    msg = "You must set 'in_vpc' to true to associate an instance with an existing ip in a vpc"
                    module.fail_json_aws(botocore.exceptions.ClientError, msg=msg)

            # Associate address object (provided or allocated) with instance
            assoc_result = associate_ip_and_device(
                ec2, module, address, private_ip_address, device_id, allow_reassociation,
                check_mode
            )
        else:
            instance = find_device(ec2, module, device_id, is_instance=False)
            # Associate address object (provided or allocated) with instance
            assoc_result = associate_ip_and_device(
                ec2, module, address, private_ip_address, device_id, allow_reassociation,
                check_mode, is_instance=False
            )

        changed = changed or assoc_result['changed']

    return {'changed': changed, 'public_ip': address['PublicIp'], 'allocation_id': address['AllocationId']}


def ensure_absent(ec2, module, address, device_id, check_mode, is_instance=True):
    if not address:
        return {'changed': False}

    # disassociating address from instance
    if device_id:
        if is_instance:
            return disassociate_ip_and_device(
                ec2, module, address, device_id, check_mode
            )
        else:
            return disassociate_ip_and_device(
                ec2, module, address, device_id, check_mode, is_instance=False
            )
    # releasing address
    else:
        return release_address(ec2, module, address, check_mode)


def allocate_address_from_pool(ec2, module, domain, check_mode, public_ipv4_pool):
    # type: (EC2Connection, str, bool, str) -> Address
    """ Overrides boto's allocate_address function to support BYOIP """
    params = {}

    if domain is not None:
        params['Domain'] = domain

    if public_ipv4_pool is not None:
        params['PublicIpv4Pool'] = public_ipv4_pool

    if check_mode:
        params['DryRun'] = 'true'

    try:
        result = ec2.allocate_address(aws_retry=True, **params)
    except (botocore.exceptions.BotoCoreError, botocore.exceptions.ClientError) as e:
        module.fail_json_aws(e, msg="Couldn't allocate Elastic IP address")
    return result


def generate_tag_dict(module, tag_name, tag_value):
    # type: (AnsibleAWSModule, str, str) -> Optional[Dict]
    """ Generates a dictionary to be passed as a filter to Amazon """
    if tag_name and not tag_value:
        if tag_name.startswith('tag:'):
            tag_name = tag_name.strip('tag:')
        return {'tag-key': tag_name}

    elif tag_name and tag_value:
        if not tag_name.startswith('tag:'):
            tag_name = 'tag:' + tag_name
        return {tag_name: tag_value}

    elif tag_value and not tag_name:
        module.fail_json(msg="parameters are required together: ('tag_name', 'tag_value')")


def main():
    argument_spec = dict(
        device_id=dict(required=False, aliases=['instance_id']),
        public_ip=dict(required=False, aliases=['ip']),
        state=dict(required=False, default='present',
                   choices=['present', 'absent']),
        in_vpc=dict(required=False, type='bool', default=False),
        reuse_existing_ip_allowed=dict(required=False, type='bool',
                                       default=False),
        release_on_disassociation=dict(required=False, type='bool', default=False),
        allow_reassociation=dict(type='bool', default=False),
        wait_timeout=dict(type='int', removed_at_date='2022-06-01', removed_from_collection='community.aws'),
        private_ip_address=dict(),
        tag_name=dict(),
        tag_value=dict(),
        public_ipv4_pool=dict()
    )

    module = AnsibleAWSModule(
        argument_spec=argument_spec,
        supports_check_mode=True,
        required_by={
            'private_ip_address': ['device_id'],
        },
    )

    ec2 = module.client('ec2', retry_decorator=AWSRetry.jittered_backoff())

    device_id = module.params.get('device_id')
    instance_id = module.params.get('instance_id')
    public_ip = module.params.get('public_ip')
    private_ip_address = module.params.get('private_ip_address')
    state = module.params.get('state')
    in_vpc = module.params.get('in_vpc')
    domain = 'vpc' if in_vpc else None
    reuse_existing_ip_allowed = module.params.get('reuse_existing_ip_allowed')
    release_on_disassociation = module.params.get('release_on_disassociation')
    allow_reassociation = module.params.get('allow_reassociation')
    tag_name = module.params.get('tag_name')
    tag_value = module.params.get('tag_value')
    public_ipv4_pool = module.params.get('public_ipv4_pool')

    if instance_id:
        warnings = ["instance_id is no longer used, please use device_id going forward"]
        is_instance = True
        device_id = instance_id
    else:
        if device_id and device_id.startswith('i-'):
            is_instance = True
        elif device_id:
            if device_id.startswith('eni-') and not in_vpc:
                module.fail_json(msg="If you are specifying an ENI, in_vpc must be true")
            is_instance = False

    tag_dict = generate_tag_dict(module, tag_name, tag_value)

    try:
        if device_id:
            address = find_address(ec2, module, public_ip, device_id, is_instance=is_instance)
        else:
            address = find_address(ec2, module, public_ip, None)

        if state == 'present':
            if device_id:
                result = ensure_present(
                    ec2, module, domain, address, private_ip_address, device_id,
                    reuse_existing_ip_allowed, allow_reassociation,
                    module.check_mode, is_instance=is_instance
                )
            else:
                if address:
                    changed = False
                else:
                    address, changed = allocate_address(
                        ec2, module, domain, reuse_existing_ip_allowed,
                        module.check_mode, tag_dict, public_ipv4_pool
                    )
                result = {
                    'changed': changed,
                    'public_ip': address['PublicIp'],
                    'allocation_id': address['AllocationId']
                }
        else:
            if device_id:
                disassociated = ensure_absent(
                    ec2, module, address, device_id, module.check_mode, is_instance=is_instance
                )

                if release_on_disassociation and disassociated['changed']:
                    released = release_address(ec2, module, address, module.check_mode)
                    result = {
                        'changed': True,
                        'disassociated': disassociated,
                        'released': released
                    }
                else:
                    result = {
                        'changed': disassociated['changed'],
                        'disassociated': disassociated,
                        'released': {'changed': False}
                    }
            else:
                released = release_address(ec2, module, address, module.check_mode)
                result = {
                    'changed': released['changed'],
                    'disassociated': {'changed': False},
                    'released': released
                }

    except (botocore.exceptions.BotoCoreError, botocore.exceptions.ClientError) as e:
        module.fail_json_aws(str(e))

    if instance_id:
        result['warnings'] = warnings
    module.exit_json(**result)


if __name__ == '__main__':
    main()



# (c) 2019 Red Hat Inc.
#
# This file is part of Ansible
#
# Ansible is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Ansible is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Ansible.  If not, see <http://www.gnu.org/licenses/>.

# Make coding more python3-ish

from __future__ import absolute_import, division, print_function

__metaclass__ = type

from ansible_collections.cisco.nxos.tests.unit.compat.mock import patch
from ansible_collections.cisco.nxos.tests.unit.modules.utils import (
    AnsibleFailJson,
)
from ansible_collections.cisco.nxos.plugins.modules import nxos_telemetry
from ansible_collections.cisco.nxos.plugins.module_utils.network.nxos.nxos import (
    NxosCmdRef,
)
from ansible_collections.cisco.nxos.plugins.module_utils.network.nxos.config.telemetry.telemetry import (
    Telemetry,
)
from .nxos_module import TestNxosModule, load_fixture, set_module_args

# TBD: These imports / import checks are only needed as a workaround for
# shippable, which fails this test due to import yaml & import ordereddict.
import pytest
from ansible_collections.cisco.nxos.plugins.module_utils.network.nxos.nxos import (
    nxosCmdRef_import_check,
)

msg = nxosCmdRef_import_check()
ignore_provider_arg = True


@pytest.mark.skipif(len(msg), reason=msg)
class TestNxosTelemetryModule(TestNxosModule):

    module = nxos_telemetry

    def setUp(self):
        super(TestNxosTelemetryModule, self).setUp()

        self.mock_FACT_LEGACY_SUBSETS = patch(
            "ansible_collections.cisco.nxos.plugins.module_utils.network.nxos.facts.facts.FACT_LEGACY_SUBSETS"
        )
        self.FACT_LEGACY_SUBSETS = self.mock_FACT_LEGACY_SUBSETS.start()

        self.mock_get_resource_connection_config = patch(
            "ansible_collections.ansible.netcommon.plugins.module_utils.network.common.cfg.base.get_resource_connection"
        )
        self.get_resource_connection_config = (
            self.mock_get_resource_connection_config.start()
        )

        self.mock_get_resource_connection_facts = patch(
            "ansible_collections.ansible.netcommon.plugins.module_utils.network.common.facts.facts.get_resource_connection"
        )
        self.get_resource_connection_facts = (
            self.mock_get_resource_connection_facts.start()
        )

        self.mock_edit_config = patch(
            "ansible_collections.cisco.nxos.plugins.module_utils.network.nxos.config.telemetry.telemetry.Telemetry.edit_config"
        )
        self.edit_config = self.mock_edit_config.start()

        self.mock_execute_show_command = patch(
            "ansible_collections.cisco.nxos.plugins.module_utils.network.nxos.nxos.NxosCmdRef.execute_show_command"
        )
        self.execute_show_command = self.mock_execute_show_command.start()

        self.mock_get_platform_shortname = patch(
            "ansible_collections.cisco.nxos.plugins.module_utils.network.nxos.nxos.NxosCmdRef.get_platform_shortname"
        )
        self.get_platform_shortname = self.mock_get_platform_shortname.start()

    def tearDown(self):
        super(TestNxosTelemetryModule, self).tearDown()
        self.mock_FACT_LEGACY_SUBSETS.stop()
        self.mock_get_resource_connection_config.stop()
        self.mock_get_resource_connection_facts.stop()
        self.mock_edit_config.stop()
        self.mock_execute_show_command.stop()
        self.get_platform_shortname.stop()

    def load_fixtures(self, commands=None, device=""):
        self.mock_FACT_LEGACY_SUBSETS.return_value = dict()
        self.get_resource_connection_config.return_value = "Connection"
        self.get_resource_connection_facts.return_value = "Connection"
        self.edit_config.return_value = None

    # ---------------------------
    # Telemetry Global Test Cases
    # ---------------------------

    def test_tms_global_merged_n9k(self):
        # Assumes feature telemetry is disabled
        # TMS global config is not present.
        self.execute_show_command.return_value = None
        self.get_platform_shortname.return_value = "N9K"
        set_module_args(
            dict(
                config=dict(
                    certificate={
                        "key": "/bootflash/sample.key",
                        "hostname": "server.example.com",
                    },
                    compression="gzip",
                    source_interface="Ethernet2/1",
                    vrf="blue",
                )
            ),
            ignore_provider_arg,
        )
        self.execute_module(
            changed=True,
            commands=[
                "feature telemetry",
                "telemetry",
                "certificate /bootflash/sample.key server.example.com",
                "destination-profile",
                "use-compression gzip",
                "source-interface Ethernet2/1",
                "use-vrf blue",
            ],
        )

    def test_tms_global_checkmode_n9k(self):
        # Assumes feature telemetry is disabled
        # TMS global config is not present.
        self.execute_show_command.return_value = None
        self.get_platform_shortname.return_value = "N9K"
        set_module_args(
            dict(
                _ansible_check_mode=True,
                config=dict(
                    certificate={
                        "key": "/bootflash/sample.key",
                        "hostname": "server.example.com",
                    },
                    compression="gzip",
                    source_interface="Ethernet2/1",
                    vrf="blue",
                ),
            ),
            ignore_provider_arg,
        )
        self.execute_module(
            changed=True,
            commands=[
                "feature telemetry",
                "telemetry",
                "certificate /bootflash/sample.key server.example.com",
                "destination-profile",
                "use-compression gzip",
                "source-interface Ethernet2/1",
                "use-vrf blue",
            ],
        )

    def test_tms_global_merged2_n9k(self):
        # Assumes feature telemetry is disabled
        # TMS global config is not present.
        # Configure only vrf
        self.execute_show_command.return_value = None
        self.get_platform_shortname.return_value = "N9K"
        set_module_args(dict(config=dict(vrf="blue")), ignore_provider_arg)
        self.execute_module(
            changed=True,
            commands=[
                "feature telemetry",
                "telemetry",
                "destination-profile",
                "use-vrf blue",
            ],
        )

    def test_tms_global_idempotent_n9k(self):
        # Assumes feature telemetry is enabled
        # TMS global config is present.
        self.execute_show_command.return_value = load_fixture(
            "nxos_telemetry", "N9K.cfg"
        )
        self.get_platform_shortname.return_value = "N9K"
        set_module_args(
            dict(
                config=dict(
                    certificate={
                        "key": "/bootflash/server.key",
                        "hostname": "localhost",
                    },
                    compression="gzip",
                    source_interface="loopback55",
                    vrf="management",
                )
            ),
            ignore_provider_arg,
        )
        self.execute_module(changed=False)

    def test_tms_global_change_cert_n9k(self):
        # Assumes feature telemetry is enabled
        # TMS global config is present
        # Change certificate
        self.execute_show_command.return_value = load_fixture(
            "nxos_telemetry", "N9K.cfg"
        )
        self.get_platform_shortname.return_value = "N9K"
        set_module_args(
            dict(
                config=dict(
                    certificate={
                        "key": "/bootflash/server.key",
                        "hostname": "my_host",
                    },
                    compression="gzip",
                    source_interface="loopback55",
                    vrf="management",
                )
            ),
            ignore_provider_arg,
        )
        self.execute_module(
            changed=True,
            commands=[
                "telemetry",
                "certificate /bootflash/server.key my_host",
            ],
        )

    def test_tms_global_change_interface_n9k(self):
        # Assumes feature telemetry is enabled
        # TMS global config is present
        # Change interface
        self.execute_show_command.return_value = load_fixture(
            "nxos_telemetry", "N9K.cfg"
        )
        self.get_platform_shortname.return_value = "N9K"
        set_module_args(
            dict(
                config=dict(
                    certificate={
                        "key": "/bootflash/server.key",
                        "hostname": "localhost",
                    },
                    compression="gzip",
                    source_interface="Ethernet8/1",
                    vrf="management",
                )
            ),
            ignore_provider_arg,
        )
        self.execute_module(
            changed=True,
            commands=[
                "telemetry",
                "destination-profile",
                "source-interface Ethernet8/1",
            ],
        )

    def test_tms_global_change_several_n9k(self):
        # Assumes feature telemetry is enabled
        # TMS global config is present
        # Change source_interface, vrf and cert
        self.execute_show_command.return_value = load_fixture(
            "nxos_telemetry", "N9K.cfg"
        )
        self.get_platform_shortname.return_value = "N9K"
        set_module_args(
            dict(
                config=dict(
                    certificate={
                        "key": "/bootflash/server_5.key",
                        "hostname": "my_host",
                    },
                    compression="gzip",
                    source_interface="Ethernet8/1",
                    vrf="blue",
                )
            ),
            ignore_provider_arg,
        )
        self.execute_module(
            changed=True,
            commands=[
                "telemetry",
                "certificate /bootflash/server_5.key my_host",
                "destination-profile",
                "source-interface Ethernet8/1",
                "use-vrf blue",
            ],
        )

    # ------------------------------
    # Telemetry DestGroup Test Cases
    # ------------------------------

    def test_tms_destgroup_input_validation_1(self):
        # Mandatory parameter 'id' missing.
        self.execute_show_command.return_value = None
        self.get_platform_shortname.return_value = "N9K"
        args = build_args(
            [
                {
                    "destination": {
                        "ip": "192.168.1.1",
                        "port": "5001",
                        "protocol": "GRPC",
                        "encoding": "GPB",
                    }
                }
            ],
            "destination_groups",
        )
        set_module_args(args, ignore_provider_arg)
        with pytest.raises(AnsibleFailJson) as errinfo:
            self.execute_module()
        testdata = errinfo.value.args[0]
        assert "Parameter <id> under <destination_groups> is required" in str(
            testdata["msg"]
        )
        assert testdata["failed"]

    def test_tms_destgroup_input_validation_2(self):
        # Parameter 'destination' is not a dict.
        self.execute_show_command.return_value = None
        self.get_platform_shortname.return_value = "N9K"
        args = build_args(
            [{"id": "88", "destination": "192.168.1.1"}], "destination_groups"
        )
        set_module_args(args, ignore_provider_arg)
        with pytest.raises(AnsibleFailJson) as errinfo:
            self.execute_module()
        testdata = errinfo.value.args[0]
        assert (
            "Parameter <destination> under <destination_groups> must be a dict"
            in str(testdata["msg"])
        )
        assert testdata["failed"]

    def test_tms_destgroup_input_validation_3(self):
        # Parameter 'destination' is not a dict.
        self.execute_show_command.return_value = None
        self.get_platform_shortname.return_value = "N9K"
        args = build_args(
            [{"id": "88", "ip": "192.168.1.1", "port": "5001"}],
            "destination_groups",
        )
        set_module_args(args, ignore_provider_arg)
        with pytest.raises(AnsibleFailJson) as errinfo:
            self.execute_module()
        testdata = errinfo.value.args[0]
        assert "Playbook entry contains unrecongnized parameters" in str(
            testdata["msg"]
        )
        assert testdata["failed"]

    def test_tms_destgroup_merged_n9k(self):
        # Assumes feature telemetry is enabled
        # TMS destgroup config is not present.
        self.execute_show_command.return_value = None
        self.get_platform_shortname.return_value = "N9K"
        args = build_args(
            [
                {
                    "id": "88",
                    "destination": {
                        "ip": "192.168.1.1",
                        "port": "5001",
                        "protocol": "GRPC",
                        "encoding": "GPB",
                    },
                },
                {
                    "id": "88",
                    "destination": {
                        "ip": "192.168.1.2",
                        "port": "6001",
                        "protocol": "GRPC",
                        "encoding": "GPB",
                    },
                },
                {
                    "id": "99",
                    "destination": {
                        "ip": "192.168.1.2",
                        "port": "6001",
                        "protocol": "GRPC",
                        "encoding": "GPB",
                    },
                },
                {
                    "id": "99",
                    "destination": {
                        "ip": "192.168.1.1",
                        "port": "5001",
                        "protocol": "GRPC",
                        "encoding": "GPB",
                    },
                },
            ],
            "destination_groups",
        )
        set_module_args(args, ignore_provider_arg)
        self.execute_module(
            changed=True,
            commands=[
                "feature telemetry",
                "telemetry",
                "destination-group 88",
                "ip address 192.168.1.1 port 5001 protocol grpc encoding gpb",
                "ip address 192.168.1.2 port 6001 protocol grpc encoding gpb",
                "destination-group 99",
                "ip address 192.168.1.2 port 6001 protocol grpc encoding gpb",
                "ip address 192.168.1.1 port 5001 protocol grpc encoding gpb",
            ],
        )

    def test_tms_destgroup_checkmode_n9k(self):
        # Assumes feature telemetry is enabled
        # TMS destgroup config is not present.
        self.execute_show_command.return_value = None
        self.get_platform_shortname.return_value = "N9K"
        args = build_args(
            [
                {
                    "id": "88",
                    "destination": {
                        "ip": "192.168.1.1",
                        "port": "5001",
                        "protocol": "GRPC",
                        "encoding": "GPB",
                    },
                }
            ],
            "destination_groups",
            state="merged",
            check_mode=True,
        )
        set_module_args(args, ignore_provider_arg)
        self.execute_module(
            changed=True,
            commands=[
                "feature telemetry",
                "telemetry",
                "destination-group 88",
                "ip address 192.168.1.1 port 5001 protocol grpc encoding gpb",
            ],
        )

    def test_tms_destgroup_merged2_n9k(self):
        # Assumes feature telemetry is enabled
        # TMS destgroup config is not present.
        # Configure only identifier
        self.execute_show_command.return_value = None
        self.get_platform_shortname.return_value = "N9K"
        args = build_args([{"id": "88"}], "destination_groups")
        set_module_args(args, ignore_provider_arg)
        self.execute_module(
            changed=True,
            commands=[
                "feature telemetry",
                "telemetry",
                "destination-group 88",
            ],
        )

    def test_tms_destgroup_idempotent_n9k(self):
        # Assumes feature telemetry is enabled
        # TMS destgroup config is not present.
        # Configure only identifier
        self.execute_show_command.return_value = load_fixture(
            "nxos_telemetry", "N9K.cfg"
        )
        self.get_platform_shortname.return_value = "N9K"
        args = build_args(
            [
                {
                    "id": "2",
                    "destination": {
                        "ip": "192.168.0.2",
                        "port": "60001",
                        "protocol": "grpc",
                        "encoding": "gpb",
                    },
                }
            ],
            "destination_groups",
        )
        set_module_args(args, ignore_provider_arg)
        self.execute_module(changed=False)

    def test_tms_destgroup_idempotent2_n9k(self):
        # Assumes feature telemetry is enabled
        # TMS destgroup config is not present.
        # Configure only identifier
        self.execute_show_command.return_value = load_fixture(
            "nxos_telemetry", "N9K.cfg"
        )
        self.get_platform_shortname.return_value = "N9K"
        args = build_args([{"id": "2"}], "destination_groups")
        set_module_args(args, ignore_provider_arg)
        self.execute_module(changed=False)

    def test_tms_destgroup_merged_aggregate_idempotent_n9k(self):
        # Assumes feature telemetry is enabled
        # TMS destgroup config is present.
        self.execute_show_command.return_value = load_fixture(
            "nxos_telemetry", "N9K.cfg"
        )
        self.get_platform_shortname.return_value = "N9K"
        args = build_args(
            [
                {
                    "id": "2",
                    "destination": {
                        "ip": "192.168.0.1",
                        "port": "50001",
                        "protocol": "gRPC",
                        "encoding": "gpb",
                    },
                },
                {
                    "id": "10",
                    "destination": {
                        "ip": "192.168.0.1",
                        "port": "50001",
                        "protocol": "gRPC",
                        "encoding": "gpb",
                    },
                },
            ],
            "destination_groups",
        )
        set_module_args(args, ignore_provider_arg)
        self.execute_module(changed=False)

    def test_tms_destgroup_change_n9k(self):
        # TMS destgroup config is not present.
        # Change protocol and encoding for dest group 2
        self.execute_show_command.return_value = load_fixture(
            "nxos_telemetry", "N9K.cfg"
        )
        self.get_platform_shortname.return_value = "N9K"
        args = build_args(
            [
                {
                    "id": "2",
                    "destination": {
                        "ip": "192.168.0.1",
                        "port": "50001",
                        "protocol": "http",
                        "encoding": "JSON",
                    },
                },
                {
                    "id": "10",
                    "destination": {
                        "ip": "192.168.0.1",
                        "port": "50001",
                        "protocol": "gRPC",
                        "encoding": "gpb",
                    },
                },
            ],
            "destination_groups",
        )
        set_module_args(args, ignore_provider_arg)
        self.execute_module(
            changed=True,
            commands=[
                "telemetry",
                "destination-group 2",
                "ip address 192.168.0.1 port 50001 protocol http encoding json",
            ],
        )

    def test_tms_destgroup_add_n9k(self):
        # TMS destgroup config is not present.
        # Add destinations to destgroup 10
        # Add new destgroup 55 and 56
        self.execute_show_command.return_value = load_fixture(
            "nxos_telemetry", "N9K.cfg"
        )
        self.get_platform_shortname.return_value = "N9K"
        args = build_args(
            [
                {
                    "id": "10",
                    "destination": {
                        "ip": "192.168.0.1",
                        "port": "50001",
                        "protocol": "gRPC",
                        "encoding": "gpb",
                    },
                },
                {
                    "id": "10",
                    "destination": {
                        "ip": "192.168.0.10",
                        "port": "50001",
                        "protocol": "gRPC",
                        "encoding": "gpb",
                    },
                },
                {
                    "id": "55",
                    "destination": {
                        "ip": "192.168.0.2",
                        "port": "50001",
                        "protocol": "gRPC",
                        "encoding": "gpb",
                    },
                },
                {"id": "56"},
            ],
            "destination_groups",
        )
        set_module_args(args, ignore_provider_arg)
        self.execute_module(
            changed=True,
            commands=[
                "telemetry",
                "destination-group 10",
                "ip address 192.168.0.10 port 50001 protocol grpc encoding gpb",
                "destination-group 55",
                "ip address 192.168.0.2 port 50001 protocol grpc encoding gpb",
                "destination-group 56",
            ],
        )

    # --------------------------------
    # Telemetry SensorGroup Test Cases
    # --------------------------------

    def test_tms_sensorgroup_merged_n9k(self):
        # Assumes feature telemetry is enabled
        # TMS sensorgroup config is not present.
        self.execute_show_command.return_value = None
        self.get_platform_shortname.return_value = "N9K"
        td55_name = (
            "sys/bgp/inst/dom-default/peer-[10.10.10.11]/ent-[10.10.10.11]"
        )
        td55_fc = (
            'or(eq(ethpmPhysIf.operSt,"down"),eq(ethpmPhysIf.operSt,"up"))'
        )
        args = build_args(
            [
                {
                    "id": "2",
                    "data_source": "NX-API",
                    "path": {
                        "name": "sys/bgp",
                        "depth": 0,
                        "query_condition": "foo",
                        "filter_condition": "foo",
                    },
                },
                {
                    "id": "2",
                    "data_source": "NX-API",
                    "path": {
                        "name": "sys/bgp/inst",
                        "depth": "unbounded",
                        "query_condition": "foo",
                        "filter_condition": "foo",
                    },
                },
                {
                    "id": "55",
                    "data_source": "DME",
                    "path": {
                        "name": td55_name,
                        "depth": 0,
                        "query_condition": "foo",
                        "filter_condition": "foo",
                    },
                },
                {
                    "id": "55",
                    "data_source": "DME",
                    "path": {
                        "name": "sys/ospf",
                        "depth": 0,
                        "query_condition": "foo",
                        "filter_condition": td55_fc,
                    },
                },
            ],
            "sensor_groups",
        )
        set_module_args(args, ignore_provider_arg)
        self.execute_module(
            changed=True,
            commands=[
                "feature telemetry",
                "telemetry",
                "sensor-group 2",
                "data-source NX-API",
                "path sys/bgp depth 0 query-condition foo filter-condition foo",
                "path sys/bgp/inst depth unbounded query-condition foo filter-condition foo",
                "sensor-group 55",
                "data-source DME",
                "path sys/bgp/inst/dom-default/peer-[10.10.10.11]/ent-[10.10.10.11] depth 0 query-condition foo filter-condition foo",
                'path sys/ospf depth 0 query-condition foo filter-condition or(eq(ethpmPhysIf.operSt,"down"),eq(ethpmPhysIf.operSt,"up"))',
            ],
        )

    def test_tms_sensorgroup_input_validation_1(self):
        # Mandatory parameter 'id' missing.
        self.execute_show_command.return_value = None
        self.get_platform_shortname.return_value = "N9K"
        args = build_args(
            [
                {
                    "data_source": "DME",
                    "path": {
                        "name": "sys/bgp",
                        "depth": 0,
                        "query_condition": "query_condition_xyz",
                        "filter_condition": "filter_condition_xyz",
                    },
                }
            ],
            "sensor_groups",
        )
        set_module_args(args, ignore_provider_arg)
        with pytest.raises(AnsibleFailJson) as errinfo:
            self.execute_module()
        testdata = errinfo.value.args[0]
        assert "Parameter <id> under <sensor_groups> is required" in str(
            testdata["msg"]
        )
        assert testdata["failed"]

    def test_tms_sensorgroup_input_validation_2(self):
        # Path present but mandatory 'name' key is not
        self.execute_show_command.return_value = None
        self.get_platform_shortname.return_value = "N9K"
        args = build_args(
            [
                {
                    "id": "77",
                    "data_source": "DME",
                    "path": {
                        "depth": 0,
                        "query_condition": "query_condition_xyz",
                        "filter_condition": "filter_condition_xyz",
                    },
                }
            ],
            "sensor_groups",
        )
        set_module_args(args, ignore_provider_arg)
        with pytest.raises(AnsibleFailJson) as errinfo:
            self.execute_module()
        testdata = errinfo.value.args[0]
        assert (
            "Parameter <path> under <sensor_groups> requires <name> key"
            in str(testdata["msg"])
        )
        assert testdata["failed"]

    def test_tms_sensorgroup_resource_key_n9k(self):
        # TMS sensorgroup config is not present.
        self.execute_show_command.return_value = None
        self.get_platform_shortname.return_value = "N9K"
        args = build_args([{"id": "77"}], "sensor_groups")
        set_module_args(args, ignore_provider_arg)
        self.execute_module(
            changed=True,
            commands=["feature telemetry", "telemetry", "sensor-group 77"],
        )

    def test_tms_sensorgroup_merged_variable_args1_n9k(self):
        # TMS sensorgroup config is not present.
        # Only path key name provided
        self.execute_show_command.return_value = None
        self.get_platform_shortname.return_value = "N9K"
        args = build_args(
            [{"id": "77", "data_source": "DME", "path": {"name": "sys/bgp"}}],
            "sensor_groups",
        )
        set_module_args(args, ignore_provider_arg)
        self.execute_module(
            changed=True,
            commands=[
                "feature telemetry",
                "telemetry",
                "sensor-group 77",
                "data-source DME",
                "path sys/bgp",
            ],
        )

    def test_tms_sensorgroup_merged_variable_args2_n9k(self):
        # TMS sensorgroup config is not present.
        # Only path keys name and depth provided
        self.execute_show_command.return_value = None
        self.get_platform_shortname.return_value = "N9K"
        args = build_args(
            [
                {
                    "id": "77",
                    "data_source": "DME",
                    "path": {"name": "sys/bgp", "depth": 0},
                }
            ],
            "sensor_groups",
        )
        set_module_args(args, ignore_provider_arg)
        self.execute_module(
            changed=True,
            commands=[
                "feature telemetry",
                "telemetry",
                "sensor-group 77",
                "data-source DME",
                "path sys/bgp depth 0",
            ],
        )

    def test_tms_sensorgroup_merged_variable_args3_n9k(self):
        # TMS sensorgroup config is not present.
        # Only path keys name, depth and query_condition provided
        self.execute_show_command.return_value = None
        self.get_platform_shortname.return_value = "N9K"
        args = build_args(
            [
                {
                    "id": "77",
                    "data_source": "DME",
                    "path": {
                        "name": "sys/bgp",
                        "depth": 0,
                        "query_condition": "query_condition_xyz",
                    },
                }
            ],
            "sensor_groups",
        )
        set_module_args(args, ignore_provider_arg)
        self.execute_module(
            changed=True,
            commands=[
                "feature telemetry",
                "telemetry",
                "sensor-group 77",
                "data-source DME",
                "path sys/bgp depth 0 query-condition query_condition_xyz",
            ],
        )

    def test_tms_sensorgroup_merged_variable_args4_n9k(self):
        # TMS sensorgroup config is not present.
        # Only path keys name, depth and filter_condition provided
        self.execute_show_command.return_value = None
        self.get_platform_shortname.return_value = "N9K"
        args = build_args(
            [
                {
                    "id": "77",
                    "data_source": "DME",
                    "path": {
                        "name": "sys/bgp",
                        "depth": 0,
                        "filter_condition": "filter_condition_xyz",
                    },
                }
            ],
            "sensor_groups",
        )
        set_module_args(args, ignore_provider_arg)
        self.execute_module(
            changed=True,
            commands=[
                "feature telemetry",
                "telemetry",
                "sensor-group 77",
                "data-source DME",
                "path sys/bgp depth 0 filter-condition filter_condition_xyz",
            ],
        )

    def test_tms_sensorgroup_merged_idempotent_n9k(self):
        # Assumes feature telemetry is enabled
        # TMS sensorgroup config is not present.
        self.execute_show_command.return_value = load_fixture(
            "nxos_telemetry", "N9K.cfg"
        )
        self.get_platform_shortname.return_value = "N9K"
        args = build_args(
            [
                {
                    "id": "2",
                    "data_source": "DME",
                    "path": {
                        "name": "sys/ospf",
                        "depth": 0,
                        "query_condition": "qc",
                        "filter_condition": "fc",
                    },
                }
            ],
            "sensor_groups",
        )
        set_module_args(args, ignore_provider_arg)
        self.execute_module(changed=False)

    def test_tms_sensorgroup_quotes_merged_idempotent_n9k(self):
        # Assumes feature telemetry is enabled
        # TMS sensorgroup config is present with quotes in NX-API path.
        self.execute_show_command.return_value = load_fixture(
            "nxos_telemetry", "N9K_SGs.cfg"
        )
        self.get_platform_shortname.return_value = "N9K"
        args = build_args(
            [
                {
                    "id": "2",
                    "data_source": "NX-API",
                    "path": {
                        "name": '"show mac address-table count"',
                        "depth": 2,
                    },
                },
                {
                    "id": "3",
                    "data_source": "NX-API",
                    "path": {"name": '"show interface ethernet1/1-52"'},
                },
                {"id": "1", "path": {"name": "sys/procsys", "depth": 1}},
            ],
            "sensor_groups",
        )
        set_module_args(args, ignore_provider_arg)
        self.execute_module(changed=False)

    def test_tms_sensorgroup_vxlan_idempotent_n9k(self):
        # TMS sensorgroup config present.
        self.execute_show_command.return_value = load_fixture(
            "nxos_telemetry", "N9K.cfg"
        )
        self.get_platform_shortname.return_value = "N9K"
        args = build_args(
            [{"id": "56", "data_source": "DME", "path": {"name": "vxlan"}}],
            "sensor_groups",
        )
        set_module_args(args, ignore_provider_arg)
        self.execute_module(changed=False)

    def test_tms_sensorgroup_idempotent_variable1_n9k(self):
        # TMS sensorgroup config is present with path key name.
        self.execute_show_command.return_value = load_fixture(
            "nxos_telemetry", "N9K.cfg"
        )
        self.get_platform_shortname.return_value = "N9K"
        args = build_args(
            [
                {
                    "id": "2",
                    "data_source": "DME",
                    "path": {
                        "name": "sys/bgp/inst/dom-default/peer-[10.10.10.11]/ent-[10.10.10.11]"
                    },
                }
            ],
            "sensor_groups",
        )
        set_module_args(args, ignore_provider_arg)
        self.execute_module(changed=False)

    def test_tms_sensorgroup_idempotent_variable2_n9k(self):
        # TMS sensorgroup config is present with path key name and depth.
        self.execute_show_command.return_value = load_fixture(
            "nxos_telemetry", "N9K.cfg"
        )
        self.get_platform_shortname.return_value = "N9K"
        args = build_args(
            [
                {
                    "id": "2",
                    "data_source": "DME",
                    "path": {"name": "boo", "depth": 0},
                }
            ],
            "sensor_groups",
        )
        set_module_args(args, ignore_provider_arg)
        self.execute_module(changed=False)

    def test_tms_sensorgroup_idempotent_resource_key_n9k(self):
        # TMS sensorgroup config is present resource key only.
        self.execute_show_command.return_value = load_fixture(
            "nxos_telemetry", "N9K.cfg"
        )
        self.get_platform_shortname.return_value = "N9K"
        args = build_args([{"id": "55"}], "sensor_groups")
        set_module_args(args, ignore_provider_arg)
        self.execute_module(changed=False)

    def test_tms_sensorgroup_present_path_environment_n9k(self):
        # TMS sensorgroup config is not present.
        # Path name 'environment' test
        self.execute_show_command.return_value = None
        self.get_platform_shortname.return_value = "N9K"
        args = build_args(
            [
                {
                    "id": "77",
                    "data_source": "YANG",
                    "path": {"name": "environment"},
                }
            ],
            "sensor_groups",
        )
        set_module_args(args, ignore_provider_arg)
        self.execute_module(
            changed=True,
            commands=[
                "feature telemetry",
                "telemetry",
                "sensor-group 77",
                "data-source YANG",
                "path environment",
            ],
        )

    def test_tms_sensorgroup_present_path_interface_n9k(self):
        # TMS sensorgroup config is not present.
        # Path name 'interface' test
        self.execute_show_command.return_value = None
        self.get_platform_shortname.return_value = "N9K"
        args = build_args(
            [
                {
                    "id": "77",
                    "data_source": "NATIVE",
                    "path": {"name": "interface"},
                }
            ],
            "sensor_groups",
        )
        set_module_args(args, ignore_provider_arg)
        self.execute_module(
            changed=True,
            commands=[
                "feature telemetry",
                "telemetry",
                "sensor-group 77",
                "data-source NATIVE",
                "path interface",
            ],
        )

    def test_tms_sensorgroup_present_path_interface_n9k(self):
        # TMS sensorgroup config is not present.
        # Path name 'resources' test
        self.execute_show_command.return_value = None
        self.get_platform_shortname.return_value = "N9K"
        args = build_args(
            [
                {
                    "id": "77",
                    "data_source": "NX-API",
                    "path": {"name": "resources"},
                }
            ],
            "sensor_groups",
        )
        set_module_args(args, ignore_provider_arg)
        self.execute_module(
            changed=True,
            commands=[
                "feature telemetry",
                "telemetry",
                "sensor-group 77",
                "data-source NX-API",
                "path resources",
            ],
        )

    # ---------------------------------
    # Telemetry Subscription Test Cases
    # ---------------------------------

    def test_tms_subscription_merged_n9k(self):
        # TMS subscription config is not present.
        self.execute_show_command.return_value = None
        self.get_platform_shortname.return_value = "N9K"
        args = build_args(
            [
                {
                    "id": 5,
                    "destination_group": 55,
                    "sensor_group": {"id": 1, "sample_interval": 1000},
                },
                {
                    "id": 88,
                    "destination_group": 3,
                    "sensor_group": {"id": 4, "sample_interval": 2000},
                },
            ],
            "subscriptions",
        )
        set_module_args(args, ignore_provider_arg)
        self.execute_module(
            changed=True,
            commands=[
                "feature telemetry",
                "telemetry",
                "subscription 5",
                "dst-grp 55",
                "snsr-grp 1 sample-interval 1000",
                "subscription 88",
                "dst-grp 3",
                "snsr-grp 4 sample-interval 2000",
            ],
        )

    def test_tms_subscription_merged_idempotent_n9k(self):
        # TMS subscription config is not present.
        self.execute_show_command.return_value = load_fixture(
            "nxos_telemetry", "N9K.cfg"
        )
        self.get_platform_shortname.return_value = "N9K"
        args = build_args(
            [
                {"id": 3},
                {
                    "id": 7,
                    "destination_group": 10,
                    "sensor_group": {"id": 2, "sample_interval": 1000},
                },
                {
                    "id": 5,
                    "destination_group": 2,
                    "sensor_group": {"id": 2, "sample_interval": 1000},
                },
            ],
            "subscriptions",
        )
        set_module_args(args, ignore_provider_arg)
        self.execute_module(changed=False)

    def test_tms_subscription_merged_change1_n9k(self):
        # TMS subscription config present.
        # Change sample interval for sensor group 2
        self.execute_show_command.return_value = load_fixture(
            "nxos_telemetry", "N9K.cfg"
        )
        self.get_platform_shortname.return_value = "N9K"
        args = build_args(
            [
                {"id": 3},
                {
                    "id": 7,
                    "destination_group": 10,
                    "sensor_group": {"id": 2, "sample_interval": 3000},
                },
                {
                    "id": 5,
                    "destination_group": 2,
                    "sensor_group": {"id": 2, "sample_interval": 1000},
                },
            ],
            "subscriptions",
        )
        set_module_args(args, ignore_provider_arg)
        self.execute_module(
            changed=True,
            commands=[
                "telemetry",
                "subscription 7",
                "snsr-grp 2 sample-interval 3000",
            ],
        )

    def test_tms_subscription_add_n9k(self):
        # TMS subscription config present.
        # Add new destination_group and sensor_group to subscription 5
        self.execute_show_command.return_value = load_fixture(
            "nxos_telemetry", "N9K.cfg"
        )
        self.get_platform_shortname.return_value = "N9K"
        args = build_args(
            [
                {"id": 3},
                {
                    "id": 7,
                    "destination_group": 10,
                    "sensor_group": {"id": 2, "sample_interval": 1000},
                },
                {
                    "id": 5,
                    "destination_group": 2,
                    "sensor_group": {"id": 2, "sample_interval": 1000},
                },
                {
                    "id": 5,
                    "destination_group": 7,
                    "sensor_group": {"id": 2, "sample_interval": 1000},
                },
                {
                    "id": 5,
                    "destination_group": 8,
                    "sensor_group": {"id": 9, "sample_interval": 1000},
                },
                {
                    "id": 5,
                    "destination_group": 9,
                    "sensor_group": {"id": 10, "sample_interval": 1000},
                },
            ],
            "subscriptions",
        )
        set_module_args(args, ignore_provider_arg)
        self.execute_module(
            changed=True,
            commands=[
                "telemetry",
                "subscription 5",
                "dst-grp 7",
                "dst-grp 8",
                "dst-grp 9",
                "snsr-grp 9 sample-interval 1000",
                "snsr-grp 10 sample-interval 1000",
            ],
        )

    def test_telemetry_full_n9k(self):
        # Assumes feature telemetry is disabled
        # TMS global config is not present.
        self.execute_show_command.return_value = None
        self.get_platform_shortname.return_value = "N9K"
        set_module_args(
            {
                "state": "merged",
                "config": {
                    "certificate": {
                        "key": "/bootflash/sample.key",
                        "hostname": "server.example.com",
                    },
                    "compression": "gzip",
                    "source_interface": "Ethernet2/1",
                    "vrf": "blue",
                    "destination_groups": [
                        {
                            "id": "88",
                            "destination": {
                                "ip": "192.168.1.1",
                                "port": "5001",
                                "protocol": "GRPC",
                                "encoding": "GPB",
                            },
                        },
                        {
                            "id": "88",
                            "destination": {
                                "ip": "192.168.1.2",
                                "port": "6001",
                                "protocol": "GRPC",
                                "encoding": "GPB",
                            },
                        },
                        {
                            "id": "99",
                            "destination": {
                                "ip": "192.168.1.2",
                                "port": "6001",
                                "protocol": "GRPC",
                                "encoding": "GPB",
                            },
                        },
                        {
                            "id": "99",
                            "destination": {
                                "ip": "192.168.1.1",
                                "port": "5001",
                                "protocol": "GRPC",
                                "encoding": "GPB",
                            },
                        },
                    ],
                    "sensor_groups": [
                        {
                            "id": "77",
                            "data_source": "DME",
                            "path": {
                                "name": "sys/bgp",
                                "depth": 0,
                                "query_condition": "query_condition_xyz",
                                "filter_condition": "filter_condition_xyz",
                            },
                        },
                        {
                            "id": "99",
                            "data_source": "DME",
                            "path": {
                                "name": "sys/bgp",
                                "depth": 0,
                                "query_condition": "query_condition_xyz",
                                "filter_condition": "filter_condition_xyz",
                            },
                        },
                    ],
                    "subscriptions": [
                        {
                            "id": 5,
                            "destination_group": 88,
                            "sensor_group": {
                                "id": 77,
                                "sample_interval": 1000,
                            },
                        },
                        {
                            "id": 5,
                            "destination_group": 99,
                            "sensor_group": {
                                "id": 77,
                                "sample_interval": 1000,
                            },
                        },
                        {
                            "id": 88,
                            "destination_group": 99,
                            "sensor_group": {
                                "id": 99,
                                "sample_interval": 2000,
                            },
                        },
                    ],
                },
            },
            ignore_provider_arg,
        )
        self.execute_module(
            changed=True,
            commands=[
                "feature telemetry",
                "telemetry",
                "certificate /bootflash/sample.key server.example.com",
                "destination-profile",
                "use-compression gzip",
                "source-interface Ethernet2/1",
                "use-vrf blue",
                "destination-group 88",
                "ip address 192.168.1.1 port 5001 protocol grpc encoding gpb",
                "ip address 192.168.1.2 port 6001 protocol grpc encoding gpb",
                "destination-group 99",
                "ip address 192.168.1.2 port 6001 protocol grpc encoding gpb",
                "ip address 192.168.1.1 port 5001 protocol grpc encoding gpb",
                "sensor-group 77",
                "data-source DME",
                "path sys/bgp depth 0 query-condition query_condition_xyz filter-condition filter_condition_xyz",
                "sensor-group 99",
                "data-source DME",
                "path sys/bgp depth 0 query-condition query_condition_xyz filter-condition filter_condition_xyz",
                "subscription 5",
                "dst-grp 88",
                "dst-grp 99",
                "snsr-grp 77 sample-interval 1000",
                "subscription 88",
                "dst-grp 99",
                "snsr-grp 99 sample-interval 2000",
            ],
        )

    def test_telemetry_deleted_input_validation_n9k(self):
        # State is 'deleted' and 'config' key present.
        self.execute_show_command.return_value = load_fixture(
            "nxos_telemetry", "N9K.cfg"
        )
        self.get_platform_shortname.return_value = "N9K"
        set_module_args(
            dict(
                state="deleted",
                config=dict(
                    certificate={
                        "key": "/bootflash/server.key",
                        "hostname": "localhost",
                    },
                    compression="gzip",
                    source_interface="loopback55",
                    vrf="management",
                ),
            ),
            ignore_provider_arg,
        )
        with pytest.raises(AnsibleFailJson) as errinfo:
            self.execute_module()
        testdata = errinfo.value.args[0]
        assert (
            "Remove config key from playbook when state is <deleted>"
            in str(testdata["msg"])
        )
        assert testdata["failed"]

    def test_telemetry_deleted_n9k(self):
        # Assumes feature telemetry is enabled
        # TMS global config is present.
        # Make absent with all playbook keys provided
        self.execute_show_command.return_value = load_fixture(
            "nxos_telemetry", "N9K.cfg"
        )
        self.get_platform_shortname.return_value = "N9K"
        set_module_args(dict(state="deleted"), ignore_provider_arg)
        self.execute_module(changed=True, commands=["no telemetry"])

    def test_telemetry_deleted_idempotent_n9k(self):
        # Assumes feature telemetry is enabled
        # TMS global config is present.
        # Make absent with all playbook keys provided
        self.execute_show_command.return_value = None
        self.get_platform_shortname.return_value = "N9K"
        set_module_args(dict(state="deleted"), ignore_provider_arg)
        self.execute_module(changed=False)

    def test_tms_replaced1_n9k(self):
        # Assumes feature telemetry is enabled
        # Modify global config and remove everything else
        self.execute_show_command.return_value = load_fixture(
            "nxos_telemetry", "N9K.cfg"
        )
        self.get_platform_shortname.return_value = "N9K"
        set_module_args(
            dict(
                state="replaced",
                config=dict(
                    certificate={
                        "key": "/bootflash/sample.key",
                        "hostname": "server.example.com",
                    },
                    compression="gzip",
                    vrf="blue",
                ),
            ),
            ignore_provider_arg,
        )
        self.execute_module(
            changed=True,
            commands=[
                "telemetry",
                "no subscription 3",
                "no subscription 4",
                "no subscription 5",
                "no subscription 6",
                "no subscription 7",
                "no sensor-group 2",
                "no sensor-group 55",
                "no sensor-group 56",
                "no destination-group 2",
                "no destination-group 10",
                "certificate /bootflash/sample.key server.example.com",
                "destination-profile",
                "no source-interface loopback55",
                "use-vrf blue",
            ],
        )

    def test_tms_replaced2_n9k(self):
        # Assumes feature telemetry is enabled
        # Remove/default all global config
        # Modify destination-group 10, add 11 and 99, remove 2
        # Modify sensor-group 55, 56
        # remove all subscriptions
        self.execute_show_command.return_value = load_fixture(
            "nxos_telemetry", "N9K.cfg"
        )
        self.get_platform_shortname.return_value = "N9K"
        set_module_args(
            {
                "state": "replaced",
                "config": {
                    "destination_groups": [
                        {
                            "id": 10,
                            "destination": {
                                "ip": "192.168.1.1",
                                "port": "5001",
                                "protocol": "GRPC",
                                "encoding": "GPB",
                            },
                        },
                        {
                            "id": 11,
                            "destination": {
                                "ip": "192.168.1.2",
                                "port": "6001",
                                "protocol": "GRPC",
                                "encoding": "GPB",
                            },
                        },
                        {
                            "id": 99,
                            "destination": {
                                "ip": "192.168.1.2",
                                "port": "6001",
                                "protocol": "GRPC",
                                "encoding": "GPB",
                            },
                        },
                        {
                            "id": "99",
                            "destination": {
                                "ip": "192.168.1.1",
                                "port": "5001",
                                "protocol": "GRPC",
                                "encoding": "GPB",
                            },
                        },
                    ],
                    "sensor_groups": [
                        {
                            "id": 55,
                            "data_source": "NX-API",
                            "path": {
                                "name": "sys/bgp",
                                "depth": 0,
                                "query_condition": "query_condition_xyz",
                                "filter_condition": "filter_condition_xyz",
                            },
                        },
                        {
                            "id": "56",
                            "data_source": "NX-API",
                            "path": {
                                "name": "sys/bgp",
                                "depth": 0,
                                "query_condition": "query_condition_xyz",
                                "filter_condition": "filter_condition_xyz",
                            },
                        },
                    ],
                },
            },
            ignore_provider_arg,
        )
        self.execute_module(
            changed=True,
            commands=[
                "telemetry",
                "no subscription 3",
                "no subscription 5",
                "no subscription 4",
                "no subscription 7",
                "no subscription 6",
                "sensor-group 56",
                "no data-source DME",
                "no path environment",
                "no path interface",
                "no path resources",
                "no path vxlan",
                "no sensor-group 2",
                "destination-group 10",
                "no ip address 192.168.0.1 port 50001 protocol grpc encoding gpb",
                "no ip address 192.168.0.2 port 60001 protocol grpc encoding gpb",
                "no destination-group 2",
                "destination-group 11",
                "ip address 192.168.1.2 port 6001 protocol grpc encoding gpb",
                "destination-group 10",
                "ip address 192.168.1.1 port 5001 protocol grpc encoding gpb",
                "destination-group 99",
                "ip address 192.168.1.2 port 6001 protocol grpc encoding gpb",
                "ip address 192.168.1.1 port 5001 protocol grpc encoding gpb",
                "sensor-group 55",
                "data-source NX-API",
                "path sys/bgp depth 0 query-condition query_condition_xyz filter-condition filter_condition_xyz",
                "sensor-group 56",
                "data-source NX-API",
                "path sys/bgp depth 0 query-condition query_condition_xyz filter-condition filter_condition_xyz",
                "no certificate /bootflash/server.key localhost",
                "no destination-profile",
            ],
        )

    def test_tms_replaced3_n9k(self):
        # Assumes feature telemetry is enabled
        # Modify vrf global config, remove default all other global config.
        # destination-group 2 destination '192.168.0.1' idempotent
        # destination-group 2 destination '192.168.0.2' remove
        # remove all other destination-groups
        # Modify sensor-group 55 and delete all others
        # Modify subscription 7, add 10 and delete all others
        self.execute_show_command.return_value = load_fixture(
            "nxos_telemetry", "N9K.cfg"
        )
        self.get_platform_shortname.return_value = "N9K"
        set_module_args(
            {
                "state": "replaced",
                "config": {
                    "vrf": "blue",
                    "destination_groups": [
                        {
                            "id": 2,
                            "destination": {
                                "ip": "192.168.0.1",
                                "port": 50001,
                                "protocol": "GRPC",
                                "encoding": "GPB",
                            },
                        }
                    ],
                    "sensor_groups": [
                        {
                            "id": 55,
                            "data_source": "NX-API",
                            "path": {
                                "name": "sys/bgp",
                                "depth": 0,
                                "query_condition": "query_condition_xyz",
                                "filter_condition": "filter_condition_xyz",
                            },
                        }
                    ],
                    "subscriptions": [
                        {
                            "id": 7,
                            "destination_group": 10,
                            "sensor_group": {
                                "id": 55,
                                "sample_interval": 1000,
                            },
                        },
                        {
                            "id": 10,
                            "destination_group": 2,
                            "sensor_group": {
                                "id": 55,
                                "sample_interval": 1000,
                            },
                        },
                    ],
                },
            },
            ignore_provider_arg,
        )
        self.execute_module(
            changed=True,
            commands=[
                "telemetry",
                "no subscription 3",
                "no subscription 5",
                "no subscription 4",
                "subscription 7",
                "no snsr-grp 2 sample-interval 1000",
                "no subscription 6",
                "no sensor-group 56",
                "no sensor-group 2",
                "no destination-group 10",
                "destination-group 2",
                "no ip address 192.168.0.2 port 60001 protocol grpc encoding gpb",
                "sensor-group 55",
                "data-source NX-API",
                "path sys/bgp depth 0 query-condition query_condition_xyz filter-condition filter_condition_xyz",
                "subscription 10",
                "dst-grp 2",
                "snsr-grp 55 sample-interval 1000",
                "subscription 7",
                "snsr-grp 55 sample-interval 1000",
                "no certificate /bootflash/server.key localhost",
                "destination-profile",
                "no use-compression gzip",
                "no source-interface loopback55",
                "use-vrf blue",
            ],
        )

    def test_tms_replaced_idempotent_n9k(self):
        # Assumes feature telemetry is enabled
        # Modify vrf global config, remove default all other global config.
        # destination-group 2 destination '192.168.0.1' idempotent
        # destination-group 2 destination '192.168.0.2' remove
        # remove all other destination-groups
        # Modify sensor-group 55 and delete all others
        # Modify subscription 7, add 10 and delete all others
        self.execute_show_command.return_value = load_fixture(
            "nxos_telemetry", "N9K.cfg"
        )
        self.get_platform_shortname.return_value = "N9K"
        set_module_args(
            {
                "state": "replaced",
                "config": {
                    "certificate": {
                        "key": "/bootflash/server.key",
                        "hostname": "localhost",
                    },
                    "compression": "gzip",
                    "vrf": "management",
                    "source_interface": "loopback55",
                    "destination_groups": [
                        {
                            "id": 2,
                            "destination": {
                                "ip": "192.168.0.1",
                                "port": 50001,
                                "protocol": "GRPC",
                                "encoding": "GPB",
                            },
                        },
                        {
                            "id": 2,
                            "destination": {
                                "ip": "192.168.0.2",
                                "port": 60001,
                                "protocol": "GRPC",
                                "encoding": "GPB",
                            },
                        },
                        {
                            "id": 10,
                            "destination": {
                                "ip": "192.168.0.1",
                                "port": 50001,
                                "protocol": "GRPC",
                                "encoding": "GPB",
                            },
                        },
                        {
                            "id": 10,
                            "destination": {
                                "ip": "192.168.0.2",
                                "port": 60001,
                                "protocol": "GRPC",
                                "encoding": "GPB",
                            },
                        },
                    ],
                    "sensor_groups": [
                        {
                            "id": 2,
                            "data_source": "DME",
                            "path": {"name": "boo", "depth": 0},
                        },
                        {
                            "id": 2,
                            "path": {
                                "name": "sys/ospf",
                                "depth": 0,
                                "query_condition": "qc",
                                "filter_condition": "fc",
                            },
                        },
                        {"id": 2, "path": {"name": "interfaces", "depth": 0}},
                        {"id": 2, "path": {"name": "sys/bgp"}},
                        {
                            "id": 2,
                            "path": {
                                "name": "sys/bgp/inst",
                                "depth": 0,
                                "query_condition": "foo",
                                "filter_condition": "foo",
                            },
                        },
                        {
                            "id": 2,
                            "path": {
                                "name": "sys/bgp/inst/dom-default/peer-[10.10.10.11]/ent-[10.10.10.11]"
                            },
                        },
                        {
                            "id": 2,
                            "path": {
                                "name": "sys/bgp/inst/dom-default/peer-[20.20.20.11]/ent-[20.20.20.11]"
                            },
                        },
                        {
                            "id": 2,
                            "path": {
                                "name": "too",
                                "depth": 0,
                                "filter_condition": "foo",
                            },
                        },
                        {"id": 55},
                        {"id": 56, "data_source": "DME"},
                        {"id": 56, "path": {"name": "environment"}},
                        {"id": 56, "path": {"name": "interface"}},
                        {"id": 56, "path": {"name": "resources"}},
                        {"id": 56, "path": {"name": "vxlan"}},
                    ],
                    "subscriptions": [
                        {"id": 3},
                        {
                            "id": 4,
                            "destination_group": 2,
                            "sensor_group": {"id": 2, "sample_interval": 1000},
                        },
                        {"id": 5, "destination_group": 2},
                        {
                            "id": 5,
                            "sensor_group": {"id": 2, "sample_interval": 1000},
                        },
                        {"id": 6, "destination_group": 10},
                        {
                            "id": 7,
                            "destination_group": 10,
                            "sensor_group": {"id": 2, "sample_interval": 1000},
                        },
                    ],
                },
            },
            ignore_provider_arg,
        )
        self.execute_module(changed=False, commands=[])


def build_args(data, type, state=None, check_mode=None):
    if state is None:
        state = "merged"
    if check_mode is None:
        check_mode = False
    args = {
        "state": state,
        "_ansible_check_mode": check_mode,
        "config": {type: data},
    }
    return args
